//(c) 2016 Dmitry Grinberg   http://dmitry.gr   me@dmitry.gr   see LICENSE.txt

#ifndef _USB_H_
#define _USB_H_


#include <avr/pgmspace.h>
#include <stdbool.h>
#include <stdint.h>



#define USB_NUM_EP		6	//number of user-usabel EPs


#define USB_EP0_SIZE		64	//so that main can have a good descriptor




#define BM_REQ_IS_DEVIC(r)		(((r) & 0x1F) == 0x00)
#define BM_REQ_IS_INTFC(r)		(((r) & 0x1F) == 0x01)
#define BM_REQ_IS_ENDPT(r)		(((r) & 0x1F) == 0x02)
#define BM_REQ_IS_OTHER(r)		(((r) & 0x1F) == 0x03)

#define BM_REQ_IS_STD(r)		(((r) & 0x60) == 0x00)
#define BM_REQ_IS_CLS(r)		(((r) & 0x60) == 0x20)
#define BM_REQ_IS_VEN(r)		(((r) & 0x60) == 0x40)



#define DESCR_TYP_DEVICE		1
#define DESCR_TYP_CONFIG		2
#define DESCR_TYP_STRING		3
#define DESCR_TYP_IFACE			4
#define DESCR_TYP_ENDPT			5
#define DESCR_TYP_DEV_QUAL		6
#define DESCR_TYP_OTH_SPD_CFG		7
#define DESCR_TYP_INT_PWR		8
#define DESCR_TYP_HID			0x21
#define DESCR_TYP_HID_REPORT		0x22
#define DESCR_TYP_HID_PHYS		0x23




void usbInit(bool lowSpeed, bool selfPowered);
void usbAttachControl(bool attach);
void usbRemoteWake(void);

#define DESCR_TYP_DEVICE		1
#define DESCR_TYP_CONFIG		2
#define DESCR_TYP_STRING		3
#define DESCR_TYP_IFACE			4
#define DESCR_TYP_ENDPT			5
#define DESCR_TYP_DEV_QUAL		6
#define DESCR_TYP_OTH_SPD_CFG		7
#define DESCR_TYP_INT_PWR		8


#define EP_TYPE_CONTROL		0b00000000
#define EP_TYPE_ISOC		0b01000000
#define EP_TYPE_BULK		0b10000000
#define EP_TYPE_INT		0b11000000

#define EP_SIZE_8		0b00000000
#define EP_SIZE_16		0b00001000
#define EP_SIZE_32		0b00010000
#define EP_SIZE_64		0b00011000
#define EP_SIZE_128		0b00100000
#define EP_SIZE_256		0b00101000
#define EP_SIZE_512		0b00110000

#define EP_DBL_BANK		0b00000010
#define EP_FLAGS_DIR_IN		0b00000001

#define EP_ENABLED		0b00000100




bool usbEpConfigure(uint8_t ep, uint8_t cfg);	//configure endpoint (or disable it). return success
void usbEpStall(uint8_t ep, bool on);		//set/clear a stall condition to be sent to host on next handshake
bool usbEpStalled(uint8_t ep);			//return if EP is stalled (we din;t get a SETUP since we stalled it)
uint8_t usbEpGetNumUsedBanks(uint8_t ep);

//for IN endpoints
bool usbEpCanSend(uint8_t ep);
void usbEpDoSendNow(uint8_t ep);	//client enqueues data, this just sets the flag to go
void usbEpDoSend(uint8_t ep, const void* dataP, uint16_t len);	//WILL send or die trying..packets multiples of EP size will send ZLP at the end

//for OUT endpoints
bool usbEpHasData(uint8_t ep);
uint16_t usbEpNumBytesAvail(uint8_t ep);			//only use if above said there is data
void usbEpRxBytes(uint8_t ep, uint8_t *buf, uint16_t num);	//read more than there is avail at your own peril


//externqal callbacks provided by the application. (all of these are clale din interrupt context, so be sensible about wasting time)

#define OFST_VAL_SIZE_H			0xFFFF	//magix ofst values to get descriptor lengths (zero for descriptors that do not exist)
#define OFST_VAL_SIZE_L			0xFFFE
uint8_t usbCbkGetDescriptor(uint8_t descrTyp, uint8_t descrIdx, uint16_t wIndex, uint16_t ofst);

#define CFG_NUM_OR_AFTER		0x80	//when called without this, just return if you cna go to this config. when called with this, configure the endpoints
bool usbCbkSetConfig(uint8_t cfg);

//called with EPNUM == 0
bool usbCbkCustomReqEp0(uint8_t bmRequestType, uint8_t bRequest, uint8_t wValueL, uint8_t wValueH, uint8_t wIndexL, uint8_t wIndexH, uint8_t wLengthL, uint8_t wLengthH);	//return success

//return true if our usb module is on and talking to master
bool usbIsConfigured(void);
uint32_t usbGetTicks(void);



//for ep0 - special
bool usbEp0canSend(void);



///descriptor helpers

#define USB_DEVICE_DESCR(nm, VID, PID, strManuf, strProd, strSer, numcfgs)	\
	static const uint8_t PROGMEM nm[] =					\
	{									\
		18,				/* bLength */			\
		DESCR_TYP_DEVICE,		/* bDescriptorType */		\
		0x00, 0x02,			/* bcdUSB (2.0) */		\
		0,				/* bDeviceClass */		\
		0,				/* bDeviceSubClass */		\
		0,				/* bDeviceProtocol */		\
		USB_EP0_SIZE,			/* bMaxPacketSize0 */		\
		((VID) & 0xFF), ((VID) >> 8),	/* idVendor */			\
		((PID) & 0xFF), ((PID) >> 8),	/* idProduct */			\
		0x00, 0x01,			/* bcdDevice */			\
		strManuf,			/* iManufacturer */		\
		strProd,			/* iProduct */			\
		strSer,				/* iSerialNumber */		\
		numcfgs				/* bNumConfigurations */	\
	};

#define USB_CONFIG_DESCR_START(nm, cfgNameStr, idx, extralen, numIfaces, attribs, milliamps)	\
	static const uint8_t PROGMEM nm[] = 							\
	{											\
		9,				/* length of descriptor in bytes */		\
		DESCR_TYP_CONFIG,		/* descriptor type */				\
		(9 + (extralen)) & 0xFF,	/* total length of data returned  */		\
		(9 + (extralen)) >>8,		/*  (incl inlined descriptors) */		\
		numIfaces,			/* number of interfaces in this config */	\
		idx,				/* index of this configuration */		\
		cfgNameStr,			/* configuration name string index */		\
		(attribs),			/* attributes */				\
		(milliamps) / 2,		/* max USB current in units of 2mA */
	
#define USB_CONFIG_DESCR_END									\
	};

//request descriptor template

#define USB_CFG_DESCR_LEN		9	//length of a config descriptor itself


#define REQ_DESCR_HANDLER_START									\
	uint8_t usbCbkGetDescriptor(uint8_t descrTyp, uint8_t descrIdx, uint16_t wIndex, uint16_t ofst){	\
		const uint8_t* dPtr;											\
		uint8_t dLenL = 0, dLenH = 0, ret;								\
		bool bRom = true;												\
		if(0){}
#define REQ_DESCR_HANDLER_END											\
		/*add else-ifs here*/											\
		else return 0;													\
																		\
		if(ofst == OFST_VAL_SIZE_H) ret = dLenH;						\
		else if(ofst == OFST_VAL_SIZE_L) ret = dLenL;					\
		else ret = bRom ? pgm_read_byte(dPtr + ofst) : dPtr[ofst];		\
		return ret;														\
	}

#define REQ_DESCR_DEVICE(dev_descr)								\
		else if(descrTyp == DESCR_TYP_DEVICE && descrIdx == 0){				\
			dPtr = dev_descr;							\
			dLenL = (sizeof(dev_descr) / sizeof(*dev_descr)) & 0xFF;		\
			dLenH = (sizeof(dev_descr) / sizeof(*dev_descr)) >> 8;			\
		}

#define REQ_DESCR_CONFIG(idx, cfg_descr)							\
		else if(descrTyp == DESCR_TYP_CONFIG && descrIdx == idx){			\
			dPtr = usbPrvDescrDeviceConfig;						\
			dLenL = (sizeof(cfg_descr) / sizeof(*cfg_descr)) & 0xFF;		\
			dLenH = (sizeof(cfg_descr) / sizeof(*cfg_descr)) >> 8;			\
		}

#define REQ_DESCR_STRINGS_START									\
		else if(descrTyp == DESCR_TYP_STRING){	/* wIndex is LanguageID */		\
			if(0){}

#define REQ_DESCR_ADD_STRING_ROM(idx, str)	REQ_DESCR_ADD_STRING(idx, str, true)
#define REQ_DESCR_ADD_STRING_RAM(idx, str)	REQ_DESCR_ADD_STRING(idx, str, false)

#define REQ_DESCR_ADD_STRING(idx, str, rom)							\
			else if(descrIdx == idx){						\
				dLenH = 0;							\
				dLenL = (sizeof(str) + 2) & 0xFF;				\
				dPtr = (const uint8_t*)&str;					\
				bRom = rom;							\
			}

#define REQ_DESCR_STRINGS_END									\
			else return 0;								\
			if(ofst == 0) return dLenL;						\
			if(ofst == 1) return DESCR_TYP_STRING;					\
			else if(ofst != OFST_VAL_SIZE_H && ofst != OFST_VAL_SIZE_L) {		\
				uint16_t val;							\
				dPtr += (ofst - 2) &~ 1;					\
				val = bRom ? pgm_read_word((uint16_t*)dPtr) : *(uint16_t*)dPtr;	\
				return (ofst & 1) ? (val >> 8) : (val & 0xFF);			\
			}									\
		}



#define USB_CUSTOM_EP0_START									\
	bool usbCbkCustomReqEp0(uint8_t bmRequestType, uint8_t bRequest, uint8_t wValueL,	\
		uint8_t wValueH, uint8_t wIndexL, uint8_t wIndexH, uint8_t wLengthL,		\
		uint8_t wLengthH){	/* return success */					\
		const void* data = NULL;							\
		const uint8_t* dataP = NULL;							\
		uint8_t sz = 0;									\
		bool sendZlpAckOnSuccess = true;						\
		(void)bmRequestType;(void)bRequest;(void)wValueL;(void)wValueH;			\
		(void)wIndexL;(void)wIndexH;(void)wLengthL;(void)wLengthH;			\
		(void)data;(void)dataP;			\
		if(0){}

#define USB_CUSTOM_EP0_END									\
		else return false;								\
		if (sendZlpAckOnSuccess)				\
			usbEpDoSend(0, data, sz);							\
		return true;									\
	}
#endif
